﻿using System;
using System.Collections.Generic;
using System.Linq.Expressions;
using System.Reflection;

namespace XCE.SelectionReset
{
    internal static class BoolExpressionParser
    {
        public static IList<ObjectDependency> Parse(Expression<Func<bool>> expression)
        {
            var visitor = new RootVisitor { Items = new List<ObjectDependency>(), RootExpression = expression };
            visitor.Visit(expression);

            return visitor.Items;
        }

        private sealed class RootVisitor : ExpressionVisitor
        {
            public LambdaExpression RootExpression { get; set; }

            public IList<ObjectDependency> Items { get; set; }

            protected override Expression VisitMember(MemberExpression node)
            {
                var property = node.Member as PropertyInfo;
                if (property != null)
                {
                    return TryReduce(property, node)
                        ?? ParseMember(node);
                }

                var field = node.Member as FieldInfo;
                if (field != null)
                {
                    return TryReduce(field, node)
                        ?? ParseMember(node);
                }

                throw new NotSupportedException();
            }

            protected override Expression VisitLambda<T>(Expression<T> node)
            {
                if (node == RootExpression)
                {
                    return base.VisitLambda(node);
                }

                return Expression.Default(node.Type);
            }

            #region NotSupportedException
            //// Not(....)
            //// protected override Expression VisitUnary(UnaryExpression node)

            //// string.IsNullOrEmpty()
            //// protected override Expression VisitMethodCall(MethodCallExpression node)

            //// protected override Expression VisitConstant(ConstantExpression node)
            //// protected override Expression VisitBinary(BinaryExpression node)
            //// protected override Expression VisitUnary(UnaryExpression node)

            protected override Expression VisitBlock(BlockExpression node)
            {
                throw new NotSupportedException();
            }

            protected override CatchBlock VisitCatchBlock(CatchBlock node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitConditional(ConditionalExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitDebugInfo(DebugInfoExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitDefault(DefaultExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitDynamic(DynamicExpression node)
            {
                throw new NotSupportedException();
            }

            protected override ElementInit VisitElementInit(ElementInit node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitGoto(GotoExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitExtension(Expression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitIndex(IndexExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitInvocation(InvocationExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitLabel(LabelExpression node)
            {
                throw new NotSupportedException();
            }

            protected override LabelTarget VisitLabelTarget(LabelTarget node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitListInit(ListInitExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitLoop(LoopExpression node)
            {
                throw new NotSupportedException();
            }

            protected override MemberAssignment VisitMemberAssignment(MemberAssignment node)
            {
                throw new NotSupportedException();
            }

            protected override MemberBinding VisitMemberBinding(MemberBinding node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitMemberInit(MemberInitExpression node)
            {
                throw new NotSupportedException();
            }

            protected override MemberListBinding VisitMemberListBinding(MemberListBinding node)
            {
                throw new NotSupportedException();
            }

            protected override MemberMemberBinding VisitMemberMemberBinding(MemberMemberBinding node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitNew(NewExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitNewArray(NewArrayExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitParameter(ParameterExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitRuntimeVariables(RuntimeVariablesExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitSwitch(SwitchExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitTry(TryExpression node)
            {
                throw new NotSupportedException();
            }

            protected override Expression VisitTypeBinary(TypeBinaryExpression node)
            {
                throw new NotSupportedException();
            }

            #endregion

            private Expression TryReduce(MemberInfo property, MemberExpression node)
            {
                // must reduce .HasValue
                if ("HasValue".Equals(property.Name, StringComparison.Ordinal)
                    && property.DeclaringType != null
                    && Nullable.GetUnderlyingType(property.DeclaringType) != null)
                {
                    // obj.HasValue => obj != default()
                    var rootVisitor = new RootVisitor { Items = Items };
                    rootVisitor.Visit(node.Expression);

                    return Expression.Default(node.Type);
                }

                return null;
            }

            private Expression ParseMember(MemberExpression node)
            {
                var visitor = new MemberVisitor();
                visitor.Parse(node);
                if (visitor.PropertyChain.Count > 0)
                {
                    Items.Add(new ObjectDependency
                    {
                        Root = visitor.Root,
                        PropertyChain = visitor.PropertyChain
                    });
                }

                return Expression.Default(node.Type);
            }
        }

        private sealed class MemberVisitor : ExpressionVisitor
        {
            private FieldInfo _rootField;

            public IList<PropertyInfo> PropertyChain { get; private set; }

            public object Root { get; private set; }

            public void Parse(Expression expression)
            {
                PropertyChain = new List<PropertyInfo>();
                Visit(expression);
            }

            protected override Expression VisitConstant(ConstantExpression node)
            {
                if (Root != null)
                {
                    throw new NotSupportedException();
                }

                Root = _rootField == null ? node.Value : _rootField.GetValue(node.Value);
                return base.VisitConstant(node);
            }

            protected override Expression VisitMember(MemberExpression node)
            {
                var property = node.Member as PropertyInfo;
                if (property != null)
                {
                    PropertyChain.Insert(0, property);
                }
                else
                {
                    if (_rootField != null)
                    {
                        throw new NotSupportedException();
                    }

                    _rootField = (FieldInfo)node.Member;
                }

                return base.VisitMember(node);
            }
        }
    }
}
