﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

namespace XCE.SelectionReset
{
    internal sealed class DependencyListener : IDisposable
    {
        private readonly IList<PropertyListener> _listeners;

        public Action Callback { get; private set; }

        public DependencyListener(Action callback)
        {
            Callback = callback;
            _listeners = new List<PropertyListener>();
        }

        public void Add(Expression<Func<bool>> expression)
        {
            var dependencies = BoolExpressionParser.Parse(expression);
            var sourceList = dependencies.Count > 1 ? Reduce(dependencies) : dependencies;

            foreach (var dependency in sourceList)
            {
                Attach(dependency.Root, dependency.PropertyChain);
            }
        }

        public void Dispose()
        {
            foreach (var i in _listeners)
            {
                i.Dispose();
            }

            _listeners.Clear();
            Callback = null;
        }

        internal static IEnumerable<ObjectDependency> Reduce(IEnumerable<ObjectDependency> sourceList)
        {
            var dependenciesByModel = new Dictionary<object, IList<ObjectDependency>>();

            foreach (var item in sourceList)
            {
                IList<ObjectDependency> dependencies;
                if (!dependenciesByModel.TryGetValue(item.Root, out dependencies))
                {
                    dependencies = new List<ObjectDependency> { item };
                    dependenciesByModel.Add(item.Root, dependencies);
                }
                else
                {
                    var flag = false;
                    foreach (var dependency in dependencies)
                    {
                        var reduced = TryReduce(dependency.PropertyChain, item.PropertyChain);
                        if (reduced != null)
                        {
                            flag = true;
                            dependency.PropertyChain = reduced;
                            break;
                        }
                    }

                    if (!flag)
                    {
                        dependencies.Add(item);
                    }
                }
            }

            return dependenciesByModel.SelectMany(i => i.Value);
        }

        internal static PropertyInfo TryGetCollectionCountProperty(Type type)
        {
            // does not expect clear generic implementations
            if (!type.IsPrimitive && typeof(IEnumerable).IsAssignableFrom(type))
            {
                return ExpandInheritance(type)
                    .Select(i => i.GetProperty("Count", BindingFlags.Public | BindingFlags.Instance | BindingFlags.DeclaredOnly))
                    .FirstOrDefault(i => i != null);
            }

            return null;
        }

        private static IEnumerable<Type> ExpandInheritance(Type root)
        {
            if (root.IsInterface)
            {
                yield return root;
                foreach (var i in root.GetInterfaces())
                {
                    yield return i;
                }
            }
            else
            {
                while (root != null && root != typeof(object))
                {
                    yield return root;
                    root = root.BaseType;
                }
            }
        }

        private static IList<PropertyInfo> TryReduce(IList<PropertyInfo> x, IList<PropertyInfo> y)
        {
            var source = x.Count < y.Count ? x : y;
            var dest = source.Equals(x) ? y : x;

            for (var i = 0; i < source.Count; i++)
            {
                if (!string.Equals(dest[i].Name, source[i].Name, StringComparison.Ordinal))
                {
                    return null;
                }
            }

            return dest;
        }

        private void ModelPropertyChanged()
        {
            Callback();
        }

        private void Attach(object model, IList<PropertyInfo> propertyChain)
        {
            var notify = model as INotifyPropertyChanged;
            if (notify == null)
            {
                throw new NotSupportedException();
            }

            var listener = new PropertyListener(notify, propertyChain, ModelPropertyChanged);
            _listeners.Add(listener);
        }

        private sealed class PropertyListener : IDisposable
        {
            private readonly PropertyInfo _modelProperty;
            private readonly IList<PropertyInfo> _childChain;
            private INotifyPropertyChanged _model;
            private Action _callback;
            private PropertyListener _childListener;

            public PropertyListener(INotifyPropertyChanged model, IList<PropertyInfo> propertyChain, Action callback)
            {
                _model = model;
                _callback = callback;
                _modelProperty = propertyChain[0];
                _childChain = propertyChain.Count > 1 ? propertyChain.Skip(1).ToArray() : null;

                // check if property is collection
                if (_childChain == null)
                {
                    var countProperty = TryGetCollectionCountProperty(_modelProperty.PropertyType);
                    if (countProperty != null)
                    {
                        _childChain = new[] { countProperty };
                    }
                }

                _model.PropertyChanged += ModelPropertyChanged;
                RebuildChildren();
            }

            public void Dispose()
            {
                _model.PropertyChanged -= ModelPropertyChanged;

                if (_childListener != null)
                {
                    _childListener.Dispose();
                    _childListener = null;
                }

                _callback = null;
                _model = null;
            }

            private void RebuildChildren()
            {
                if (_childChain != null)
                {
                    if (_childListener != null)
                    {
                        _childListener.Dispose();
                        _childListener = null;
                    }

                    var model = _modelProperty.GetValue(_model, new object[0]) as INotifyPropertyChanged;
                    if (model != null)
                    {
                        _childListener = new PropertyListener(model, _childChain, _callback);
                    }
                }
            }

            private void ModelPropertyChanged(object sender, PropertyChangedEventArgs e)
            {
                if (string.Equals(_modelProperty.Name, e.PropertyName, StringComparison.OrdinalIgnoreCase))
                {
                    RebuildChildren();
                    _callback();
                }
            }
        }
    }
}