﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows;

namespace XCE.SelectionReset
{
    internal sealed class RecipeViewModel : ViewModelBase
    {
        private ObservableCollection<LayerBaseViewModel> _layers;

        private LayerBaseViewModel _selectedLayer;

        private bool _isLayerDefinitionFinished = true;

        public ObservableCollection<LayerBaseViewModel> Layers
        {
            get { return _layers; }
            private set
            {
                if (_layers == value) return;
                _layers = value;
                OnPropertyChanged();
            }
        }

        public LayerBaseViewModel SelectedLayer
        {
            get { return _selectedLayer; }
            set
            {
                _selectedLayer = value;
                OnPropertyChanged();
            }
        }

        public bool IsLayerDefinitionFinished
        {
            get { return _isLayerDefinitionFinished; }
            private set
            {
                _isLayerDefinitionFinished = value;
                OnPropertyChanged();
            }
        }

        public RecipeViewModel()
        {
            Commands.Add("AddLayerA", AddLayerA);
            Commands.Add("AddLayerB", AddLayerB);
            Commands.Add("AddLayerC", AddLayerC);
            Commands.Add<LayerBaseViewModel>("DeleteLayer", DeleteLayer);
            Commands.Add("MoveLayerUp", MoveLayerUp, () => SelectedLayer != null && SelectedLayer.Order < Layers.Count);
            Commands.Add("MoveLayerDown", MoveLayerDown, () => SelectedLayer != null && SelectedLayer.Order > 1);
            Commands.Add("FinishLayerDefinition", FinishLayerDefinition);
            Commands.Add("ReopenLayerDefinition", ReopenLayerDefinition);
            Initialize();
            
        }

        private void Initialize()
        {
            Layers = new ObservableCollection<LayerBaseViewModel>();
            IsLayerDefinitionFinished = false;
            AddLayerA();
            SelectedLayer = Layers.First();
        }

        private void AddLayerA()
        {
            var newLayerName = GenerateLayerName(LayerType.A);
            var newLayerModel = new LayerA
            {
                Id = SequentialGuid.NewGuid(),
                Name = newLayerName,
                Order = Layers.Count + 1
            };
            var newLayer = new LayerAViewModel(newLayerModel);
            Layers.Insert(0, newLayer);
            SelectedLayer = newLayer;
        }

        private void AddLayerB()
        {
            var newLayerName = GenerateLayerName(LayerType.B);
            var newLayerModel = new LayerB
            {
                Id = SequentialGuid.NewGuid(),
                Name = newLayerName,
                Order = Layers.Count + 1,
            };
            var newLayer = new LayerBViewModel(newLayerModel, Layers);
            Layers.Insert(0, newLayer);
            SelectedLayer = newLayer;
        }

        private void AddLayerC()
        {
            var newLayerName = GenerateLayerName(LayerType.C);
            var newLayerModel = new LayerC
            {
                Id = SequentialGuid.NewGuid(),
                Name = newLayerName,
                Order = Layers.Count + 1
            };
            var newLayer = new LayerCViewModel(newLayerModel, Layers);
            Layers.Insert(0, newLayer);
            SelectedLayer = newLayer;
        }

        private void DeleteLayer(LayerBaseViewModel layer)
        {
            var answer = MessageBox.Show(
                $"Are you sure want to delete layer '{layer.Name}'?\n\n" +
                "It will be also deleted from all other layers’ targets.",
                "Delete layer",
                MessageBoxButton.YesNo,
                MessageBoxImage.Question);
            if (answer != MessageBoxResult.Yes)
            {
                return;
            }

            Layers.Remove(layer);

            foreach (var otherLayer in Layers)
            {
                switch (otherLayer.Type)
                {
                    case LayerType.A:
                        break;
                    case LayerType.B:
                        var layerBViewModel = otherLayer as LayerBViewModel;
                        var removedLayers = layerBViewModel.TargetLayers.First(targetLayer => ((TargetLayerViewModel)targetLayer).Id == layer.Id);
                        layerBViewModel.TargetLayers.Remove(removedLayers);
                        break;
                    case LayerType.C:
                        var layerCViewModel = otherLayer as LayerCViewModel;
                        var removedLayer1 = layerCViewModel.LayersSet1.First(targetLayer => ((TargetLayerViewModel)targetLayer).Id == layer.Id);
                        layerCViewModel.LayersSet1.Remove(removedLayer1);
                        var removedLayer2 = layerCViewModel.LayersSet2.First(targetLayer => ((TargetLayerViewModel)targetLayer).Id == layer.Id);
                        layerCViewModel.LayersSet2.Remove(removedLayer2);
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }

            if (Layers.Count == 0)
            {
                AddLayerA();
            }

            RebuildOrder();
        }

        private void MoveLayerUp()
        {
            var nextLayer = Layers.FirstOrDefault(layer => layer.Order == SelectedLayer.Order + 1);
            SwitchLayers(SelectedLayer, nextLayer);
        }

        private void MoveLayerDown()
        {
            var prevLayer = Layers.FirstOrDefault(layer => layer.Order == SelectedLayer.Order - 1);
            SwitchLayers(SelectedLayer, prevLayer);
        }

        private void SwitchLayers(LayerBaseViewModel layerA, LayerBaseViewModel layerB)
        {
            var sourceIndex = Layers.IndexOf(layerA);
            var targetIndex = Layers.IndexOf(layerB);

            Layers.Move(sourceIndex, targetIndex);

            if (SelectedLayer == layerA)
            {
                SelectedLayer = null;
                SelectedLayer = layerA;
            }
            else if (SelectedLayer == layerB)
            {
                SelectedLayer = null;
                SelectedLayer = layerB;
            }

            RebuildOrder();
        }

        private void RebuildOrder()
        {
            for (var i = 0; i < Layers.Count; i++)
            {
                Layers[i].Order = Layers.Count - i;
            }

            foreach (var layer in Layers)
            {
                switch (layer.Type)
                {
                    case LayerType.A:
                        break;
                    case LayerType.B:
                        var layerBViewModel = layer as LayerBViewModel;
                        var removedLayer = layerBViewModel.TargetLayers.FirstOrDefault(targetLayer => ((TargetLayerViewModel)targetLayer).Order > layerBViewModel.Order);
                        layerBViewModel.TargetLayers.Remove(removedLayer);
                        break;
                    case LayerType.C:
                        var layerCViewModel = layer as LayerCViewModel;
                        var removedLayer1 = layerCViewModel.LayersSet1.FirstOrDefault(targetLayer => ((TargetLayerViewModel)targetLayer).Order > layerCViewModel.Order);
                        layerCViewModel.LayersSet1.Remove(removedLayer1);
                        var removedLayer2 = layerCViewModel.LayersSet2.FirstOrDefault(targetLayer => ((TargetLayerViewModel)targetLayer).Order > layerCViewModel.Order);
                        layerCViewModel.LayersSet2.Remove(removedLayer2);
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }
        }

        private string GenerateLayerName(LayerType layerType)
        {
            var no = Layers.Count(layer => layer.Type == layerType);
            while (true)
            {
                var name = $"{layerType} {++no}";
                if (Layers.All(layer => !layer.Name.EqualsIgnoreCase(name)))
                {
                    return name;
                }
            }
        }

        private void FinishLayerDefinition()
        {
            IsLayerDefinitionFinished = true;
        }

        private void ReopenLayerDefinition()
        {
            var answer = MessageBox.Show(
                "All settings in the Measurements tab and the Rules tab will be lost.\n\n" +
                "Would you like to proceed?",
                "Reopen layer definition",
                MessageBoxButton.YesNo,
                MessageBoxImage.Question);
            if (answer != MessageBoxResult.No)
            {
                IsLayerDefinitionFinished = false;
            }
        }
    }
}